/************************************************************************/
/* Disassembler for KINT virtual processor				*/
/************************************************************************/

#include "kint.hpp"

/* Register operand */
static void
reg_op( int op, int regNum )
{
  cmd.Operands[op].type = o_reg;
  cmd.Operands[op].reg = regNum;
}

/* Register bit operand */
static void
regbit_op( int op, int regNum, int bit )
{
  cmd.Operands[op].type = o_reg_bit;
  cmd.Operands[op].reg = regNum;
  cmd.Operands[op].v_bit = bit;
}

/* Immediate operand */
static void
imm_op( int op, int offset, ulong value, char dtyp )
{
  cmd.Operands[op].type = o_imm;
  cmd.Operands[op].value = value;
  cmd.Operands[op].dtyp = dtyp;
  cmd.Operands[op].offb = offset;
}

/* Phrase operand */
static void
phrase_op( int op, char dtyp, int reg1, int reg2 = -1 )
{
  cmd.Operands[op].type = o_phrase;
  cmd.Operands[op].dtyp = dtyp;
  cmd.Operands[op].phrase = reg1;
  cmd.Operands[op].specval = reg2;
}

/* Displacement operand */
static void
disp_op( int op, char dtyp, int offset, int regNum = -1, ea_t mem = 0 )
{
  cmd.Operands[op].type = o_displ;
  cmd.Operands[op].dtyp = dtyp;
  cmd.Operands[op].addr = offset;
  cmd.Operands[op].phrase = regNum;
  cmd.Operands[op].value = mem;
}

/* Address operand */
static void
addr_op( int op, int offset, ea_t address )
{
  cmd.Operands[op].type = o_mem;
  cmd.Operands[op].addr = address;
  cmd.Operands[op].offb = offset;
}

/* Code operand (e.g. JP destination) */
static void
code_op( int op, int offset, ea_t address )
{
  cmd.Operands[op].type = o_near;
  cmd.Operands[op].addr = address;
  cmd.Operands[op].dtyp = dt_dword;
  cmd.Operands[op].offb = offset;
}

static optype_t
size_reg ( int reg ) {
  if( reg < 0x80 )
    return dt_dword;
  if( reg < 0xe0 )
    return dt_word;
  return dt_byte;
}

//----------------------------------------------------------------------
// analyze an basic instruction
static int
ana_basic( void )
{
  // get the command code byte
  ushort	val, code = ua_next_word(  );
  uchar		op = code >> 8;
  uchar		dr = code & 0xff;
  uchar		reg, regh;
  uint		opcode;

  switch ( op )
  {
    case 0x00:			// NOP
      cmd.itype = KINT_NOP;
      break;
    case 0x7D:			// SAF b
    case 0x7E:			// REF b
    case 0x85:			// INC b
    case 0x86:			// INC w
    case 0x87:			// INC d
    case 0x88:			// DEC b
    case 0x89:			// DEC w
    case 0x8A:			// DEC d
    case 0x8B:			// NOT b
    case 0x8C:			// NOT w
    case 0x8D:			// NOT d
    case 0x8E:			// NEG b
    case 0x8F:			// NEG w
    case 0x90:			// NEG d
    case 0x91:			// PUSH b
    case 0x92:			// PUSH w
    case 0x93:			// PUSH d
    case 0x94:			// POP b
    case 0x95:			// POP w
    case 0x96:			// POP d
    case 0x97:			// CBW w
    case 0xB8:			// CBD d
    case 0xB9:			// CWD d
      reg_op( 0, dr );
      switch ( op )
      {
	case 0x7D:
	  cmd.itype = KINT_SAF; break;
	case 0x7E:
	  cmd.itype = KINT_REF; break;
	case 0x85: case 0x86: case 0x87:
	  cmd.itype = KINT_INC; break;
	case 0x88: case 0x89: case 0x8A:
	  cmd.itype = KINT_DEC; break;
	case 0x8B: case 0x8C: case 0x8D:
	  cmd.itype = KINT_NOT; break;
	case 0x8E: case 0x8F: case 0x90:
	  cmd.itype = KINT_NEG; break;
	case 0x91: case 0x92: case 0x93:
	  cmd.itype = KINT_PUSH; break;
	case 0x94: case 0x95: case 0x96:
	  cmd.itype = KINT_POP; break;
	case 0x97:
	  cmd.itype = KINT_CBW; break;
	case 0xB8:
	  cmd.itype = KINT_CBD; break;
	case 0xB9:
	  cmd.itype = KINT_CWD; break;
      }
      break;
    case 0x03:			// SHL b,b
    case 0x04:			// SHL w,b
    case 0x05:			// SHL d,b
    case 0x06:			// SHR b,b
    case 0x07:			// SHR w,b
    case 0x08:			// SHR d,b
    case 0x09:			// ROL b,b
    case 0x0A:			// ROL w,b
    case 0x0B:			// ROL d,b
    case 0x0C:			// ROR b,b
    case 0x0D:			// ROR w,b
    case 0x0E:			// ROR d,b
    case 0x0F:			// ADD b,b
    case 0x10:			// ADD w,w
    case 0x11:			// ADD d,d
    case 0x12:			// SUB b,b
    case 0x13:			// SUB w,w
    case 0x14:			// SUB d,d
    case 0x15:			// MUL b,b
    case 0x16:			// MUL w,w
    case 0x17:			// MUL d,d
    case 0x18:			// DIV b,b
    case 0x19:			// DIV w,w
    case 0x1A:			// DIV d,d
    case 0x1B:			// MOD b,b
    case 0x1C:			// MOD w,w
    case 0x1D:			// MOD d,d
    case 0x1E:			// OR b,b
    case 0x1F:			// OR w,w
    case 0x20:			// OR d,d
    case 0x21:			// AND b,b
    case 0x22:			// AND w,w
    case 0x23:			// AND d,d
    case 0x24:			// XOR b,b
    case 0x25:			// XOR w,w
    case 0x26:			// XOR d,d
    case 0x27:			// CMP b,b
    case 0x28:			// CMP w,w
    case 0x29:			// CMP d,d
    case 0x2A:			// TST b,b
    case 0x2B:			// TST w,w
    case 0x2C:			// TST d,d
    case 0x2D:			// XCH b,b
    case 0x2E:			// XCH w,w
    case 0x2F:			// XCH d,d
    case 0x30:			// LDR b,b
    case 0x31:			// LDR w,w
    case 0x32:			// LDR d,d
    case 0x33:			// LDE d,b
    case 0x34:			// LDE d,w
    case 0x35:			// LDE w,b
    case 0x36:			// ABIT w,b
    case 0x37:			// ABIT b,b
      reg_op( 0, dr );
      reg_op( 1, ua_next_word(	) & 0xff );
      switch ( op )
      {
	case 0x03: case 0x04: case 0x05:
	  cmd.itype = KINT_SHL; break;
	case 0x06: case 0x07: case 0x08:
	  cmd.itype = KINT_SHR; break;
	case 0x09: case 0x0A: case 0x0B:
	  cmd.itype = KINT_ROL; break;
	case 0x0C: case 0x0D: case 0x0E:
	  cmd.itype = KINT_ROR; break;
	case 0x0F: case 0x10: case 0x11:
	  cmd.itype = KINT_ADD; break;
	case 0x12: case 0x13: case 0x14:
	  cmd.itype = KINT_SUB; break;
	case 0x15: case 0x16: case 0x17:
	  cmd.itype = KINT_MUL; break;
	case 0x18: case 0x19: case 0x1A:
	  cmd.itype = KINT_DIV; break;
	case 0x1B: case 0x1C: case 0x1D:
	  cmd.itype = KINT_MOD; break;
	case 0x1E: case 0x1F: case 0x20:
	  cmd.itype = KINT_OR; break;
	case 0x21: case 0x22: case 0x23:
	  cmd.itype = KINT_AND; break;
	case 0x24: case 0x25: case 0x26:
	  cmd.itype = KINT_XOR; break;
	case 0x27: case 0x28: case 0x29:
	  cmd.itype = KINT_CMP; break;
	case 0x2A: case 0x2B: case 0x2C:
	  cmd.itype = KINT_TST; break;
	case 0x2D: case 0x2E: case 0x2F:
	  cmd.itype = KINT_XCH; break;
	case 0x30: case 0x31: case 0x32:
	  cmd.itype = KINT_LDR; break;
	case 0x33: case 0x34: case 0x35:
	  cmd.itype = KINT_LDE; break;
	case 0x36: case 0x37:
	  cmd.itype = KINT_ABIT; break;
      }
      break;
    case 0xBD:			// SHL b,=wval
    case 0xBE:			// SHL w,=wval
    case 0xBF:			// SHL d,=wval
    case 0xC0:			// SHR b,=wval
    case 0xC1:			// SHR w,=wval
    case 0xC2:			// SHR d,=wval
    case 0xC3:			// ROL b,=wval
    case 0xC4:			// ROL w,=wval
    case 0xC5:			// ROL d,=wval
    case 0xC6:			// ROR b,=wval
    case 0xC7:			// ROR w,=wval
    case 0xC8:			// ROR d,=wval
    case 0xA1:			// LDS b,=dval
    case 0xA2:			// LDS w,=dval
    case 0xA3:			// LDS d,=dval
    case 0xAD:			// STS b,=dval
    case 0xAE:			// STS w,=dval
    case 0xAF:			// STS d,=dval
      reg_op( 0, dr );
      imm_op( 1, 2, ua_next_word(  ), dt_word );
      switch ( op )
      {
	case 0xBD: case 0xBE: case 0xBF:
	  cmd.itype = KINT_SHL; break;
	case 0xC0: case 0xC1: case 0xC2:
	  cmd.itype = KINT_SHR; break;
	case 0xC3: case 0xC4: case 0xC5:
	  cmd.itype = KINT_ROL; break;
	case 0xC6: case 0xC7: case 0xC8:
	  cmd.itype = KINT_ROR; break;
	case 0xA1: case 0xA2: case 0xA3:
	  cmd.itype = KINT_LDS; break;
	case 0xAD: case 0xAE: case 0xAF:
	  cmd.itype = KINT_STS; break;
      }
      break;
    case 0x38:			// LD0 b,d
    case 0x39:			// LD0 w,d
    case 0x3A:			// LD0 d,d
    case 0x3B:			// LD1 b,d
    case 0x3C:			// LD1 w,d
    case 0x3D:			// LD1 d,d
    case 0x3E:			// LD2 b,d
    case 0x3F:			// LD2 w,d
    case 0x40:			// LD2 d,d
    case 0x41:			// LDS b,d
    case 0x42:			// LDS w,d
    case 0x43:			// LDS d,d
    case 0x44:			// ST0 b,d
    case 0x45:			// ST0 w,d
    case 0x46:			// ST0 d,d
    case 0x47:			// ST1 b,d
    case 0x48:			// ST1 w,d
    case 0x49:			// ST1 d,d
    case 0x4A:			// ST2 b,d
    case 0x4B:			// ST2 w,d
    case 0x4C:			// ST2 d,d
    case 0x4D:			// STS b,d
    case 0x4E:			// STS w,d
    case 0x4F:			// STS d,d
    case 0x50:			// PU0 b,d
    case 0x51:			// PU0 w,d
    case 0x52:			// PU0 d,d
    case 0x53:			// PO0 b,d
    case 0x54:			// PO0 w,d
    case 0x55:			// PO0 d,d
      reg_op( 0, dr );
      reg_op( 1, ua_next_word(	) & 0xff );
      switch ( op )
      {
	case 0x38: case 0x39: case 0x3A:
	  cmd.itype = KINT_LD0; break;
	case 0x3B: case 0x3C: case 0x3D:
	  cmd.itype = KINT_LD1; break;
	case 0x3E: case 0x3F: case 0x40:
	  cmd.itype = KINT_LD2; break;
	case 0x41: case 0x42: case 0x43:
	  cmd.itype = KINT_LDS; break;
	case 0x44: case 0x45: case 0x46:
	  cmd.itype = KINT_ST0; break;
	case 0x47: case 0x48: case 0x49:
	  cmd.itype = KINT_ST1; break;
	case 0x4A: case 0x4B: case 0x4C:
	  cmd.itype = KINT_ST2; break;
	case 0x4D: case 0x4E: case 0x4F:
	  cmd.itype = KINT_STS; break;
	case 0x50: case 0x51: case 0x52:
	  cmd.itype = KINT_PU0; break;
	case 0x53: case 0x54: case 0x55:
	  cmd.itype = KINT_PO0; break;
      }
      break;
    case 0x98:			// LD0 b,=dval
    case 0x99:			// LD0 w,=dval
    case 0x9A:			// LD0 d,=dval
    case 0x9B:			// LD1 b,=dval
    case 0x9C:			// LD1 w,=dval
    case 0x9D:			// LD1 d,=dval
    case 0x9E:			// LD2 b,=dval
    case 0x9F:			// LD2 w,=dval
    case 0xA0:			// LD2 d,=dval
    case 0xA4:			// ST0 b,=dval
    case 0xA5:			// ST0 w,=dval
    case 0xA6:			// ST0 d,=dval
    case 0xA7:			// ST1 b,=dval
    case 0xA8:			// ST1 w,=dval
    case 0xA9:			// ST1 d,=dval
    case 0xAA:			// ST2 b,=dval
    case 0xAB:			// ST2 w,=dval
    case 0xAC:			// ST2 d,=dval
      reg_op( 0, dr );
      imm_op( 1, 2, ua_next_long(  ), dt_dword );
      switch ( op )
      {
	case 0x98: case 0x99: case 0x9A:
	  cmd.itype = KINT_LD0; break;
	case 0x9B: case 0x9C: case 0x9D:
	  cmd.itype = KINT_LD1; break;
	case 0x9E: case 0x9F: case 0xA0:
	  cmd.itype = KINT_LD2; break;
	case 0xA4: case 0xA5: case 0xA6:
	  cmd.itype = KINT_ST0; break;
	case 0xA7: case 0xA8: case 0xA9:
	  cmd.itype = KINT_ST1; break;
	case 0xAA: case 0xAB: case 0xAC:
	  cmd.itype = KINT_ST2; break;
      }
      break;
    case 0xB0:			// IJMP b,=dval
    case 0xB1:			// IJMP w,=dval
    case 0xB2:			// IJMP d,=dval
    case 0xB3:			// KJMP b,=dval
    case 0xB4:			// KJMP w,=dval
    case 0xB5:			// KJMP d,=dval
      reg_op( 0, dr );
      addr_op( 1, 2, ua_next_long(  ) );
      switch ( op )
      {
	case 0xB0: case 0xB1: case 0xB2:
	  cmd.itype = KINT_IJMP; break;
	case 0xB3: case 0xB4: case 0xB5:
	  cmd.itype = KINT_KJMP; break;
      }
      break;
    case 0xBA:			// LDI b,=wval
    case 0xBB:			// LDI w,=wval
    case 0xBC:			// LDI d,=dval
    case 0xC9:			// ADD b,=val
    case 0xCA:			// ADD w,=val
    case 0xCB:			// ADD d,=val
    case 0xCC:			// SUB b,=val
    case 0xCD:			// SUB w,=val
    case 0xCE:			// SUB d,=val
    case 0xCF:			// MUL b,=val
    case 0xD0:			// MUL w,=val
    case 0xD1:			// MUL d,=val
    case 0xD2:			// DIV b,=val
    case 0xD3:			// DIV w,=val
    case 0xD4:			// DIV d,=val
    case 0xD5:			// MOD b,=val
    case 0xD6:			// MOD w,=val
    case 0xD7:			// MOD d,=val
    case 0xD8:			// OR b,=val
    case 0xD9:			// OR w,=val
    case 0xDA:			// OR d,=val
    case 0xDB:			// AND b,=val
    case 0xDC:			// AND w,=val
    case 0xDD:			// AND d,=val
    case 0xDE:			// XOR b,=val
    case 0xDF:			// XOR w,=val
    case 0xE0:			// XOR d,=val
    case 0xE1:			// CMP b,=val
    case 0xE2:			// CMP w,=val
    case 0xE3:			// CMP d,=val
    case 0xE4:			// TST b,=val
    case 0xE5:			// TST w,=val
    case 0xE6:			// TST d,=val
      reg_op( 0, dr );
      switch ( op )
      {
	case 0xBA: case 0xC9: case 0xCC: case 0xCF:
	case 0xD2: case 0xD5: case 0xD8: case 0xDB:
	case 0xDE: case 0xE1: case 0xE4:	// b,=val
	  imm_op( 1, 2, ua_next_word(  ), dt_byte );
	  break;
	case 0xBB: case 0xCA: case 0xCD: case 0xD0:
	case 0xD3: case 0xD6: case 0xD9: case 0xDC:
	case 0xDF: case 0xE2: case 0xE5:	// w,=val
	  imm_op( 1, 2, ua_next_word(  ), dt_word );
	  break;
	case 0xBC: case 0xCB: case 0xCE: case 0xD1:
	case 0xD4: case 0xD7: case 0xDA: case 0xDD:
	case 0xE0: case 0xE3: case 0xE6:	// d,=val
	  imm_op( 1, 2, ua_next_long(  ), dt_dword );
	  break;
      }
      switch ( op )
      {
	case 0xBA: case 0xBB: case 0xBC:
	  cmd.itype = KINT_LDI; break;
	case 0xC9: case 0xCA: case 0xCB:
	  cmd.itype = KINT_ADD; break;
	case 0xCC: case 0xCD: case 0xCE:
	  cmd.itype = KINT_SUB; break;
	case 0xCF: case 0xD0: case 0xD1:
	  cmd.itype = KINT_MUL; break;
	case 0xD2: case 0xD3: case 0xD4:
	  cmd.itype = KINT_DIV; break;
	case 0xD5: case 0xD6: case 0xD7:
	  cmd.itype = KINT_MOD; break;
	case 0xD8: case 0xD9: case 0xDA:
	  cmd.itype = KINT_OR; break;
	case 0xDB: case 0xDC: case 0xDD:
	  cmd.itype = KINT_AND; break;
	case 0xDE: case 0xDF: case 0xE0:
	  cmd.itype = KINT_XOR; break;
	case 0xE1: case 0xE2: case 0xE3:
	  cmd.itype = KINT_CMP; break;
	case 0xE4: case 0xE5: case 0xE6:
	  cmd.itype = KINT_TST; break;
      }
      break;
    case 0xE8:			// JE +offset (0 to 255 WORDS from CP)
    case 0xE9:			// JE -offset (-256 to -1 WORDS from CP)
    case 0xEA:			// JNE +offset
    case 0xEB:			// JNE -offset
    case 0xEC:			// JL +offset
    case 0xED:			// JL -offset
    case 0xEE:			// JLE +offset
    case 0xEF:			// JLE -offset
    case 0xF0:			// JG +offset
    case 0xF1:			// JG -offset
    case 0xF2:			// JGE +offset
    case 0xF3:			// JGE -offset
    case 0xF4:			// JEV +offset
    case 0xF5:			// JEV -offset
    case 0xF6:			// JOD +offset
    case 0xF7:			// JOD -offset
    case 0xF8:			// JMP +offset
    case 0xF9:			// JMP -offset
      if ( op & 1 )
	code_op( 0, 2, cmd.ip + cmd.size + ( 0xffffff00 | dr ) * 2 );
      else
	code_op( 0, 2, cmd.ip + cmd.size + dr * 2 );
      switch ( op )
      {
	case 0xE8: case 0xE9: cmd.itype = KINT_JE;  break;
	case 0xEA: case 0xEB: cmd.itype = KINT_JNE; break;
	case 0xEC: case 0xED: cmd.itype = KINT_JL;  break;
	case 0xEE: case 0xEF: cmd.itype = KINT_JLE; break;
	case 0xF0: case 0xF1: cmd.itype = KINT_JG;  break;
	case 0xF2: case 0xF3: cmd.itype = KINT_JGE; break;
	case 0xF4: case 0xF5: cmd.itype = KINT_JEV; break;
	case 0xF6: case 0xF7: cmd.itype = KINT_JOD; break;
	case 0xF8: case 0xF9: cmd.itype = KINT_JMP; break;
      }
      break;
    case 0x78:			// CLR b,b
    case 0x79:			// PUR b,b
    case 0x7A:			// POR b,b
      reg_op( 0, dr );
      reg_op( 1, ua_next_word(	) & 0xff );
      switch ( op )
      {
	case 0x78: cmd.itype = KINT_CLR; break;
	case 0x79: cmd.itype = KINT_PUR; break;
	case 0x7A: cmd.itype = KINT_POR; break;
      }
      break;
    case 0x58:			// LD0 b,d,=dval
    case 0x59:			// LD0 w,d,=dval
    case 0x5A:			// LD0 d,d,=dval
    case 0x5B:			// LD1 b,d,=dval
    case 0x5C:			// LD1 w,d,=dval
    case 0x5D:			// LD1 d,d,=dval
    case 0x5E:			// LD2 b,d,=dval
    case 0x5F:			// LD2 w,d,=dval
    case 0x60:			// LD2 d,d,=dval
    case 0x61:			// LDS b,d,=dval
    case 0x62:			// LDS w,d,=dval
    case 0x63:			// LDS d,d,=dval
    case 0x64:			// ST0 b,d,=dval
    case 0x65:			// ST0 w,d,=dval
    case 0x66:			// ST0 d,d,=dval
    case 0x67:			// ST1 b,d,=dval
    case 0x68:			// ST1 w,d,=dval
    case 0x69:			// ST1 d,d,=dval
    case 0x6A:			// ST2 b,d,=dval
    case 0x6B:			// ST2 w,d,=dval
    case 0x6C:			// ST2 d,d,=dval
    case 0x6D:			// STS b,d,=dval
    case 0x6E:			// STS w,d,=dval
    case 0x6F:			// STS d,d,=dval
    case 0x70:			// PU0 b,d,=dval
    case 0x71:			// PU0 w,d,=dval
    case 0x72:			// PU0 d,d,=dval
    case 0x73:			// PO0 b,d,=dval
    case 0x74:			// PO0 w,d,=dval
    case 0x75:			// PO0 d,d,=dval
      reg_op( 0, dr );
      reg_op( 1, ua_next_word(	) & 0xff );
      imm_op( 2, 4, ua_next_long(  ), dt_dword );
      switch ( op )
      {
	case 0x58: case 0x59: case 0x5A:
	  cmd.itype = KINT_LD0; break;
	case 0x5B: case 0x5C: case 0x5D:
	  cmd.itype = KINT_LD1; break;
	case 0x5E: case 0x5F: case 0x60:
	  cmd.itype = KINT_LD2; break;
	case 0x61: case 0x62: case 0x63:
	  cmd.itype = KINT_LDS; break;
	case 0x64: case 0x65: case 0x66:
	  cmd.itype = KINT_ST0; break;
	case 0x67: case 0x68: case 0x69:
	  cmd.itype = KINT_ST1; break;
	case 0x6A: case 0x6B: case 0x6C:
	  cmd.itype = KINT_ST2; break;
	case 0x6D: case 0x6E: case 0x6F:
	  cmd.itype = KINT_STS; break;
	case 0x70: case 0x71: case 0x72:
	  cmd.itype = KINT_PU0; break;
	case 0x73: case 0x74: case 0x75:
	  cmd.itype = KINT_PO0; break;
      }
      break;
    case 0x76:			// LSA d,=dval
      reg_op( 0, dr );
      imm_op( 1, 2, ua_next_word(  ), dt_word );
      cmd.itype = KINT_LSA;
      break;
    case 0x77:			// B2I r,r (byte-2-index)
      reg_op( 0, dr );
      reg_op( 1, ua_next_word(	) & 0xff );
      cmd.itype = KINT_B2I;
      break;
    case 0xE7:			// JMP =val (long)
      code_op( 0, 2, ua_next_long(  ) );
      cmd.itype = KINT_LJMP;
      break;
    case 0xB7:			// JSI d,=dval
      reg_op( 0, dr );
      imm_op( 1, 2, ua_next_long(  ), dt_dword );
      cmd.itype = KINT_JSI;
      break;
    case 0xFC:			// JSI =dval
      addr_op( 0, 2, ua_next_long(  ) );
      cmd.itype = KINT_JSI;
      break;
    case 0xFA:			// SysFunc
      cmd.itype = SYS_FILESIZE + dr;
      break;
    case 0xB6:			// JSR d,=dval
      reg_op( 0, dr );
      imm_op( 1, 2, ua_next_long(  ), dt_dword );
      cmd.itype = KINT_JSR;
      break;
    case 0xFB:			// JSR =dval , LJxx =dval , JMP[]
      if ( get_byte( 6 ) < 3 )	// reset 'dr' if code assembled for < 3.0 KINT
	dr = 0;
      switch ( dr )
      {
	case 0x00:		// JSR =dval
	  cmd.itype = KINT_JSR;
	  code_op( 0, 2, ua_next_long(	) );
	  break;
	case 0x02:		// JMP [reg]
	  phrase_op( 0, dt_dword, ua_next_word(	 ) & 0xff );
	  cmd.itype = KINT_JMP;
	  break;
	case 0x03:		// JMP [reg+off]
	  reg = ua_next_word(  ) & 0xff;
	  disp_op( 0, dt_dword, ua_next_long(  ), reg );
	  cmd.itype = KINT_JMP;
	  break;
	case 0x04:		// JMP [off]
	  disp_op( 0, dt_dword, ua_next_long(  ) );
	  cmd.itype = KINT_JMP;
	  break;
	case 0x05:		// JMP [reg+off+off]
	  reg = ua_next_word(  ) & 0xff;
	  disp_op( 0, dt_dword, ua_next_long(  ), reg, ua_next_long(  ) );
	  cmd.itype = KINT_JMP;
	  break;
	case 0x06:		// JMP [reg+reg]
	  phrase_op( 0, dt_byte, ua_next_byte(	), ua_next_byte(  ) );
	  cmd.itype = KINT_JMP;
	  break;
	case 0x08:		// LJE
	case 0x09:		// LJNE
	case 0x0A:		// LJL
	case 0x0B:		// LJLE
	case 0x0C:		// LJG
	case 0x0D:		// LJGE
	case 0x0E:		// LJEV
	case 0x0F:		// LJOD
	  code_op( 0, 2, ua_next_long(	) );
	  switch ( dr )
	  {
	    case 0x08: cmd.itype = KINT_LJE;  break;
	    case 0x09: cmd.itype = KINT_LJNE; break;
	    case 0x0A: cmd.itype = KINT_LJL;  break;
	    case 0x0B: cmd.itype = KINT_LJLE; break;
	    case 0x0C: cmd.itype = KINT_LJG;  break;
	    case 0x0D: cmd.itype = KINT_LJGE; break;
	    case 0x0E: cmd.itype = KINT_LJEV; break;
	    case 0x0F: cmd.itype = KINT_LJOD; break;
	  }
      }
      break;
    case 0x7F:			// RET b,=0
    case 0x80:			// RET w,=0
    case 0x81:			// RET d,=0
    case 0x82:			// RET b,=1
    case 0x83:			// RET w,=1
    case 0x84:			// RET d,=1
      reg_op( 0, dr );
      imm_op( 1, 0, op < 0x82 ? 0 : 1, dt_byte );
      // fall through
    case 0xFD:			// RET
      cmd.itype = KINT_RET;
      break;
    case 0xFE:			// misc
      uchar reg;
      switch ( dr )
      {
	case 0x0004:		// AFR
	  cmd.itype = KINT_AFR;
	  imm_op( 0, 2, ua_next_word(  ), dt_word );
	  break;
	case 0x0000:		// BRK
	case 0x0001:		// NOP
	case 0x0002:		// TERM
	case 0x0003:		// HALT
	case 0x0005:		// DFR
	case 0x0006:		// USM
	case 0x0007:		// SM
	case 0x006B:		// RETC
	case 0x0088:		// ADF
	case 0x0089:		// SUF
	case 0x008A:		// MUF
	case 0x008B:		// DIF
	case 0x008C:		// CMF
	case 0x008D:		// NEF
	  switch ( dr )
	  {
            case 0x00: cmd.itype = KINT_BRK;  break;
            case 0x01: cmd.itype = KINT_NOP;  break;
            case 0x02: cmd.itype = KINT_TERM; break;
            case 0x03: cmd.itype = KINT_HALT; break;
            case 0x05: cmd.itype = KINT_DFR;  break;
            case 0x06: cmd.itype = KINT_USM;  break;
            case 0x07: cmd.itype = KINT_SM;   break;
            case 0x6B: cmd.itype = KINT_RETC; break;
            case 0x88: cmd.itype = KINT_ADF;  break;
            case 0x89: cmd.itype = KINT_SUF;  break;
            case 0x8A: cmd.itype = KINT_MUF;  break;
            case 0x8B: cmd.itype = KINT_DIF;  break;
            case 0x8C: cmd.itype = KINT_CMF;  break;
            case 0x8D: cmd.itype = KINT_NEF;  break;
	  }
	  break;
	case 0x0010:		// LDCM b,b
	case 0x0011:		// LDCM w,b
	case 0x0012:		// LDCM d,b
	case 0x0013:		// LDC b,b
	case 0x0014:		// LDC w,b
	case 0x0015:		// LDC d,b
	case 0x0016:		// LDCP b,b
	case 0x0017:		// LDCP w,b
	case 0x0018:		// LDCP d,b
	case 0x0019:		// STCM b,b
	case 0x001A:		// STCM w,b
	case 0x001B:		// STCM d,b
	case 0x001C:		// STC b,b
	case 0x001D:		// STC w,b
	case 0x001E:		// STC d,b
	case 0x001F:		// STCP b,b
	case 0x0020:		// STCP w,b
	case 0x0021:		// STCP d,b
	  reg_op( 1, ua_next_byte(  ) );
	  reg_op( 0, ua_next_byte(  ) );
	  switch ( dr )
	  {
	    case 0x10: case 0x11: case 0x12:
	      cmd.itype = KINT_LDCM; break;
	    case 0x13: case 0x14: case 0x15:
	      cmd.itype = KINT_LDC;  break;
	    case 0x16: case 0x17: case 0x18:
	      cmd.itype = KINT_LDCP; break;
	    case 0x19: case 0x1A: case 0x1B:
	      cmd.itype = KINT_STCM; break;
	    case 0x1C: case 0x1D: case 0x1E:
	      cmd.itype = KINT_STC;  break;
	    case 0x1F: case 0x20: case 0x21:
	      cmd.itype = KINT_STCP; break;
	  }
	  break;
	case 0x0025:		// PUSH [reg]B
	case 0x0026:		// PUSH [reg+off]B
	case 0x0027:		// PUSH [mem]B
	case 0x0028:		// PUSH [reg+mem+off]B
	case 0x0029:		// PUSH [reg+reg]B
	case 0x002A:		// PUSH [reg]W
	case 0x002B:		// PUSH [reg+off]W
	case 0x002C:		// PUSH [mem]W
	case 0x002D:		// PUSH [reg+mem+off]W
	case 0x002E:		// PUSH [reg+reg]W
	case 0x002F:		// PUSH [reg]D
	case 0x0030:		// PUSH [reg+off]D
	case 0x0031:		// PUSH [mem]D
	case 0x0032:		// PUSH [reg+mem+off]D
	case 0x0033:		// PUSH [reg+reg]D
	case 0x0034:		// POP [reg]B
	case 0x0035:		// POP [reg+off]B
	case 0x0036:		// POP [mem]B
	case 0x0037:		// POP [reg+mem+off]B
	case 0x0038:		// POP [reg+reg]B
	case 0x0039:		// POP [reg]W
	case 0x003A:		// POP [reg+off]W
	case 0x003B:		// POP [mem]W
	case 0x003C:		// POP [reg+mem+off]W
	case 0x003D:		// POP [reg+reg]W
	case 0x003E:		// POP [reg]D
	case 0x003F:		// POP [reg+off]D
	case 0x0040:		// POP [mem]D
	case 0x0041:		// POP [reg+mem+off]D
	case 0x0042:		// POP [reg+reg]D
	case 0x0043:		// INC [reg]B
	case 0x0044:		// INC [reg+off]B
	case 0x0045:		// INC [mem]B
	case 0x0046:		// INC [reg+mem+off]B
	case 0x0047:		// INC [reg+reg]B
	case 0x0048:		// INC [reg]W
	case 0x0049:		// INC [reg+off]W
	case 0x004A:		// INC [mem]W
	case 0x004B:		// INC [reg+mem+off]W
	case 0x004C:		// INC [reg+reg]W
	case 0x004D:		// INC [reg]D
	case 0x004E:		// INC [reg+off]D
	case 0x004F:		// INC [mem]D
	case 0x0050:		// INC [reg+mem+off]D
	case 0x0051:		// INC [reg+reg]D
	case 0x0052:		// DEC [reg]B
	case 0x0053:		// DEC [reg+off]B
	case 0x0054:		// DEC [mem]B
	case 0x0055:		// DEC [reg+mem+off]B
	case 0x0056:		// DEC [reg+reg]B
	case 0x0057:		// DEC [reg]W
	case 0x0058:		// DEC [reg+off]W
	case 0x0059:		// DEC [mem]W
	case 0x005A:		// DEC [reg+mem+off]W
	case 0x005B:		// DEC [reg+reg]W
	case 0x005C:		// DEC [reg]D
	case 0x005D:		// DEC [reg+off]D
	case 0x005E:		// DEC [mem]D
	case 0x005F:		// DEC [reg+mem+off]D
	case 0x0060:		// DEC [reg+reg]D
	case 0x0078:		// PUF [reg]D
	case 0x0079:		// PUF [reg+off]D
	case 0x007A:		// PUF [mem]D
	case 0x007B:		// PUF [reg+mem+off]D
	case 0x007C:		// PUF [reg+reg]D
	case 0x0080:		// POF [reg]D
	case 0x0081:		// POF [reg+off]D
	case 0x0082:		// POF [mem]D
	case 0x0083:		// POF [reg+mem+off]D
	case 0x0084:		// POF [reg+reg]D
	case 0x00AA:		// NEG [reg]B
	case 0x00AB:		// NEG [reg+off]B
	case 0x00AC:		// NEG [mem]B
	case 0x00AD:		// NEG [reg+mem+off]B
	case 0x00AE:		// NEG [reg+reg]B
	case 0x00AF:		// NEG [reg]W
	case 0x00B0:		// NEG [reg+off]W
	case 0x00B1:		// NEG [mem]W
	case 0x00B2:		// NEG [reg+mem+off]W
	case 0x00B3:		// NEG [reg+reg]W
	case 0x00B4:		// NEG [reg]D
	case 0x00B5:		// NEG [reg+off]D
	case 0x00B6:		// NEG [mem]D
	case 0x00B7:		// NEG [reg+mem+off]D
	case 0x00B8:		// NEG [reg+reg]D
	case 0x00B9:		// NOT [reg]B
	case 0x00BA:		// NOT [reg+off]B
	case 0x00BB:		// NOT [mem]B
	case 0x00BC:		// NOT [reg+mem+off]B
	case 0x00BD:		// NOT [reg+reg]B
	case 0x00BE:		// NOT [reg]W
	case 0x00BF:		// NOT [reg+off]W
	case 0x00C0:		// NOT [mem]W
	case 0x00C1:		// NOT [reg+mem+off]W
	case 0x00C2:		// NOT [reg+reg]W
	case 0x00C3:		// NOT [reg]D
	case 0x00C4:		// NOT [reg+off]D
	case 0x00C5:		// NOT [mem]D
	case 0x00C6:		// NOT [reg+mem+off]D
	case 0x00C7:		// NOT [reg+reg]D
	case 0x00C8:		// TST [reg]B
	case 0x00C9:		// TST [reg+off]B
	case 0x00CA:		// TST [mem]B
	case 0x00CB:		// TST [reg+mem+off]B
	case 0x00CC:		// TST [reg+reg]B
	case 0x00CD:		// TST [reg]W
	case 0x00CE:		// TST [reg+off]W
	case 0x00CF:		// TST [mem]W
	case 0x00D0:		// TST [reg+mem+off]W
	case 0x00D1:		// TST [reg+reg]W
	case 0x00D2:		// TST [reg]D
	case 0x00D3:		// TST [reg+off]D
	case 0x00D4:		// TST [mem]D
	case 0x00D5:		// TST [reg+mem+off]D
	case 0x00D6:		// TST [reg+reg]D
	  {
	    switch ( dr )
	    {
	      case 0x25: case 0x34: case 0x43: case 0x52: case 0xAA:
	      case 0xB9: case 0xC8:				// [reg]B
		phrase_op( 0, dt_byte, ua_next_word(  ) & 0xff );
		break;
	      case 0x2A: case 0x39: case 0x48: case 0x57: case 0xAF:
	      case 0xBE: case 0xCD:				// [reg]W
		phrase_op( 0, dt_word, ua_next_word(  ) & 0xff );
		break;
	      case 0x2F: case 0x3E: case 0x4D: case 0x5C: case 0x78:
	      case 0x80: case 0xB4: case 0xC3: case 0xD2:	// [reg]D
		phrase_op( 0, dt_dword, ua_next_word(  ) & 0xff );
		break;
	      case 0x29: case 0x38: case 0x47: case 0x56: case 0xAE:
	      case 0xBD: case 0xCC:				// [reg+reg]B
		phrase_op( 0, dt_byte, ua_next_byte(  ), ua_next_byte(	) );
		break;
	      case 0x2E: case 0x3D: case 0x4C: case 0x5B: case 0xB3:
	      case 0xC2: case 0xD1:				// [reg+reg]W
		phrase_op( 0, dt_word, ua_next_byte(  ), ua_next_byte(	) );
		break;
	      case 0x33: case 0x42: case 0x51: case 0x60: case 0x7C:
	      case 0x84: case 0xB8: case 0xC7: case 0xD6:	// [reg+reg]D
		phrase_op( 0, dt_dword, ua_next_byte(  ), ua_next_byte(	 ) );
		break;
	      case 0x27: case 0x36: case 0x45: case 0x54: case 0xAC:
	      case 0xBB: case 0xCA:				// [mem]B
		disp_op( 0, dt_byte, ua_next_long(  ) );
		break;
	      case 0x2C: case 0x3B: case 0x4A: case 0x59: case 0xB1:
	      case 0xC0: case 0xCF:				// [mem]W
		disp_op( 0, dt_word, ua_next_long(  ) );
		break;
	      case 0x31: case 0x40: case 0x4F: case 0x5E: case 0x7A:
	      case 0x82: case 0xB6: case 0xC5: case 0xD4:	// [mem]D
		disp_op( 0, dt_dword, ua_next_long(  ) );
		break;
	      case 0x26: case 0x35: case 0x44: case 0x53: case 0xAB:
	      case 0xBA: case 0xC9:				// [reg+off]B
		reg = ua_next_word(  ) & 0xff;
		disp_op( 0, dt_byte, ua_next_long(  ), reg );
		break;
	      case 0x2B: case 0x3A: case 0x49: case 0x58: case 0xB0:
	      case 0xBF: case 0xCE:				// [reg+off]W
		reg = ua_next_word(  ) & 0xff;
		disp_op( 0, dt_word, ua_next_long(  ), reg );
		break;
	      case 0x30: case 0x3F: case 0x4E: case 0x5D: case 0x79:
	      case 0x81: case 0xB5: case 0xC4: case 0xD3:	// [reg+off]D
		reg = ua_next_word(  ) & 0xff;
		disp_op( 0, dt_dword, ua_next_long(  ), reg );
		break;
	      case 0x28: case 0x37: case 0x46: case 0x55: case 0xAD:
	      case 0xBC: case 0xCB:				// [reg+mem+off]B
		reg = ua_next_word(  ) & 0xff;
		disp_op( 0, dt_byte, ua_next_long(  ), reg,
		    ua_next_long(  ) );
		break;
	      case 0x2D: case 0x3C: case 0x4B: case 0x5A: case 0xB2:
	      case 0xC1: case 0xD0:				// [reg+mem+off]W
		reg = ua_next_word(  ) & 0xff;
		disp_op( 0, dt_word, ua_next_long(  ), reg,
		    ua_next_long(  ) );
		break;
	      case 0x32: case 0x41: case 0x50: case 0x5F: case 0x7B:
	      case 0x83: case 0xB7: case 0xC6: case 0xD5:	// [reg+mem+off]D
		reg = ua_next_word(  ) & 0xff;
		disp_op( 0, dt_dword, ua_next_long(  ), reg,
		    ua_next_long(  ) );
		break;
	    }
	    switch ( dr )
	    {
	      case 0x25: case 0x26: case 0x27: case 0x28: case 0x29:
	      case 0x2A: case 0x2B: case 0x2C: case 0x2D: case 0x2E:
	      case 0x2F: case 0x30: case 0x31: case 0x32: case 0x33:	// PUSH
		cmd.itype = KINT_PUSH;
		break;
	      case 0x34: case 0x35: case 0x36: case 0x37: case 0x38:
	      case 0x39: case 0x3A: case 0x3B: case 0x3C: case 0x3D:
	      case 0x3E: case 0x3F: case 0x40: case 0x41: case 0x42:	// POP
		cmd.itype = KINT_POP;
		break;
	      case 0x43: case 0x44: case 0x45: case 0x46: case 0x47:
	      case 0x48: case 0x49: case 0x4A: case 0x4B: case 0x4C:
	      case 0x4D: case 0x4E: case 0x4F: case 0x50: case 0x51:	// INC
		cmd.itype = KINT_INC;
		break;
	      case 0x52: case 0x53: case 0x54: case 0x55: case 0x56:
	      case 0x57: case 0x58: case 0x59: case 0x5A: case 0x5B:
	      case 0x5C: case 0x5D: case 0x5E: case 0x5F: case 0x60:	// DEC
		cmd.itype = KINT_DEC;
		break;
	      case 0xAA: case 0xAB: case 0xAC: case 0xAD: case 0xAE:
	      case 0xAF: case 0xB0: case 0xB1: case 0xB2: case 0xB3:
	      case 0xB4: case 0xB5: case 0xB6: case 0xB7: case 0xB8:	// NEG
		cmd.itype = KINT_NEG;
		break;
	      case 0xB9: case 0xBA: case 0xBB: case 0xBC: case 0xBD:
	      case 0xBE: case 0xBF: case 0xC0: case 0xC1: case 0xC2:
	      case 0xC3: case 0xC4: case 0xC5: case 0xC6: case 0xC7:	// NOT
		cmd.itype = KINT_NOT;
		break;
	      case 0xC8: case 0xC9: case 0xCA: case 0xCB: case 0xCC:
	      case 0xCD: case 0xCE: case 0xCF: case 0xD0: case 0xD1:
	      case 0xD2: case 0xD3: case 0xD4: case 0xD5: case 0xD6:	// TST
		cmd.itype = KINT_TST;
		break;
	      case 0x78: case 0x79: case 0x7A: case 0x7B: case 0x7C:	// PUF
		cmd.itype = KINT_PUF;
		break;
	      case 0x80: case 0x81: case 0x82: case 0x83: case 0x84:	// POF
		cmd.itype = KINT_POF;
		break;
	    }
	  }
	  break;
	case 0x0022:		// PUB =imm
	  imm_op( 0, 2, ua_next_word(  ), dt_word );
	  cmd.itype = KINT_PUB;
	  break;
	case 0x0023:		// PUW =imm
	  imm_op( 0, 2, ua_next_word(  ), dt_word );
	  cmd.itype = KINT_PUW;
	  break;
	case 0x0024:		// PUD =imm
	  imm_op( 0, 2, ua_next_long(  ), dt_dword );
	  cmd.itype = KINT_PUD;
	  break;
	case 0x007E:		// PUF =imm
	  imm_op( 0, 2, ua_next_long(  ), dt_dword );
	  cmd.itype = KINT_PUF;
	  break;
	case 0x007F:		// PUF =0
	  imm_op( 0, 0, 0, dt_dword );
	  cmd.itype = KINT_PUF;
	  break;
	case 0x0086:		// POF 1 (discard)
	  imm_op( 0, 0, 1, dt_dword );
	  cmd.itype = KINT_POF;
	  break;
	case 0x0087:		// POF 2 (discard)
	  imm_op( 0, 0, 2, dt_dword );
	  cmd.itype = KINT_POF;
	  break;
	case 0x0061:		// PUSC b
	case 0x0062:		// PUSC w
	case 0x0063:		// PUSC d
	  reg_op( 0, ua_next_word(  ) & 0xff );
	  cmd.itype = KINT_PUSC;
	  break;
	case 0x0064:		// POPC b
	case 0x0065:		// POPC w
	case 0x0066:		// POPC d
	  reg_op( 0, ua_next_word(  ) & 0xff );
	  cmd.itype = KINT_POPC;
	  break;
	case 0x0067:		// JSRC =dval
	  code_op( 0, 2, ua_next_long(	) );
	  cmd.itype = KINT_JSRC;
	  break;
	case 0x0068:		// JSIC =dval
	  addr_op( 0, 2, ua_next_long(	) );
	  cmd.itype = KINT_JSIC;
	  break;
	case 0x0069:		// JSRC d,=dval
	  reg_op( 0, ua_next_word(  ) & 0xff );
	  imm_op( 1, 4, ua_next_long(  ), dt_dword );
	  cmd.itype = KINT_JSRC;
	  break;
	case 0x006A:		// JSIC d,=dval
	  reg_op( 0, ua_next_word(  ) & 0xff );
	  imm_op( 1, 4, ua_next_long(  ), dt_dword );
	  cmd.itype = KINT_JSIC;
	  break;
	case 0x0071:		// FADD r,r
	case 0x0072:		// FSUB r,r
	case 0x0073:		// FMUL r,r
	case 0x0074:		// FDIV r,r
	case 0x0077:		// FCMP r,r
	  reg_op( 0, ua_next_byte(  ) );
	  reg_op( 1, ua_next_byte(  ) );
	  switch ( dr )
	  {
	    case 0x71: cmd.itype = KINT_FADD; break;
	    case 0x72: cmd.itype = KINT_FSUB; break;
	    case 0x73: cmd.itype = KINT_FMUL; break;
	    case 0x74: cmd.itype = KINT_FDIV; break;
	    case 0x77: cmd.itype = KINT_FCMP; break;
	  }
	  break;
	case 0x0070:		// FNEG r
	case 0x0075:		// FF2I r
	case 0x0076:		// FI2F r
	case 0x007D:		// PUF r
	case 0x0085:		// POF r
	case 0x008E:		// I2F r
	case 0x008F:		// F2I r
	  reg_op( 0, ua_next_word(  ) & 0xff );
	  switch ( dr )
	  {
	    case 0x70: cmd.itype = KINT_FNEG; break;
	    case 0x75: cmd.itype = KINT_FF2I; break;
	    case 0x76: cmd.itype = KINT_FI2F; break;
	    case 0x7D: cmd.itype = KINT_PUF;  break;
	    case 0x85: cmd.itype = KINT_POF;  break;
	    case 0x8E: cmd.itype = KINT_I2F;  break;
	    case 0x8F: cmd.itype = KINT_F2I;  break;
	  }
	  break;
      }
      break;
    case 0x56:			// memory ops
    case 0x57:			// memory ops
      {
	uchar	      chunk = ( code >> 7 ) & 3;
	uchar	      inst = ( code >> 4 ) & 7;
	uchar	      amode = ( code >> 0 ) & 0xf;
	if ( amode == 0xf )
	  switch ( inst )
	  {
	    case 0:		// [reg]
	      if ( chunk )	// pea
	      {
	        phrase_op( 0, dt_dword, ua_next_word( ) & 0xff );
	      }
	      else		// lea
	      {
		reg = ua_next_byte(  );
		reg_op( 0, reg );
	        phrase_op( 1, size_reg( reg ), ua_next_byte(  ) ); 
	      }
	      break;
	    case 1:		// [reg+off] 
	      if ( chunk )	// pea
	      {
		reg = ua_next_word( ) & 0xff;
	        disp_op( 0, dt_dword, ua_next_long(  ), reg );
	      }
	      else		// lea
	      {
		reg = ua_next_byte(  );
		reg_op( 0, reg );
		regh = ua_next_byte( );
	        disp_op( 1, size_reg( reg ), ua_next_long(  ), regh );
	      }
	      break;
	    case 2:		// [mem]
	      if ( chunk )	// pea
	      {
	        disp_op( 0, dt_dword, ua_next_long(  ) );
	      }
	      else		// lea
	      {
		reg = ua_next_word(  ) & 0xff;
		reg_op( 0, reg );
	        disp_op( 1, size_reg( reg ), ua_next_long(  ) );
	      }
	      break;
	    case 3:		// [reg+mem+off] 
	      if ( chunk )	// pea
	      {
		reg = ua_next_word( ) & 0xff;
	        disp_op( 0, dt_dword, ua_next_long(  ), reg,
		         ua_next_long(  ) );
	      }
	      else		// lea
	      {
		reg = ua_next_byte(  );
		reg_op( 0, reg );
		regh = ua_next_byte( );
	        disp_op( 1, size_reg( reg ), ua_next_long(  ), regh,
		   ua_next_long(  ) );
	      }
	      break;
	    case 4:		// [reg+reg] 
	      if ( chunk )	// pea
	      {
	        phrase_op( 0, dt_dword, ua_next_byte(  ), ua_next_byte(  ) );
	      }
	      else		// lea
	      {
		reg = ua_next_byte(  );
		regh = ua_next_byte(  );
		reg_op( 0, reg );
	        phrase_op( 1, size_reg( reg ), regh, ua_next_word(  ) & 0xff );
	      }
	      break;
	  }
	else
	  switch ( amode )
	  {
	    case 0:		// regl,[regh]
	      reg_op( 0, ua_next_byte(	) );
	      phrase_op( 1, chunk, ua_next_byte(  ) );
	      break;
	    case 1:		// regl,[regh+off]
	      reg_op( 0, ua_next_byte(	) );
	      reg = ua_next_byte(  );
	      disp_op( 1, chunk, ua_next_long(	), reg );
	      break;
	    case 2:		// regl,[mem]
	      reg_op( 0, ua_next_word(	) & 0xff );
	      disp_op( 1, chunk, ua_next_long(	) );
	      break;
	    case 3:		// regl,[regh+mem+off]
	      reg_op( 0, ua_next_byte(	) );
	      regh = ua_next_byte(  );
	      disp_op( 1, chunk, ua_next_long(	), regh, 
		  ua_next_long(  ) );
	      break;
	    case 4:		// regl,[regh+reg2]
	      reg_op( 0, ua_next_byte(	) );
	      regh = ua_next_byte(  );
	      phrase_op( 1, chunk, regh, ua_next_word( ) & 0xff );
	      break;
	    case 5:		// [regl],regh
	      phrase_op( 0, chunk, ua_next_byte(  ) );
	      reg_op( 1, ua_next_byte(	) );
	      break;
	    case 6:		// [regl],imm
	      phrase_op( 0, chunk, ua_next_word(  ) & 0xff );
	      if ( chunk <= 1 )
		imm_op( 1, 4, ua_next_word(  ), dt_word );
	      else
		imm_op( 1, 4, ua_next_long(  ), dt_dword );
	      break;
	    case 7:		// [regl+off],regh
	      reg = ua_next_byte(  );
	      regh = ua_next_byte(  );
	      disp_op( 0, chunk, ua_next_long(	), reg );
	      reg_op( 1, regh );
	      break;
	    case 8:		// [regl+off],imm
	      reg = ua_next_word(  ) & 0xff;
	      disp_op( 0, chunk, ua_next_long(	), reg );
	      if ( chunk <= 1 )
		imm_op( 1, 8, ua_next_word(  ), dt_word );
	      else
		imm_op( 1, 8, ua_next_long(  ), dt_dword );
	      break;
	    case 9:		// [mem],regl
	      reg = ua_next_word(  ) & 0xff;
	      disp_op( 0, chunk, ua_next_long(	) );
	      reg_op( 1, reg );
	      break;
	    case 10:		// [mem],imm
	      disp_op( 0, chunk, ua_next_long(	) );
	      if ( chunk <= 1 )
		imm_op( 1, 6, ua_next_word(  ), dt_word );
	      else
		imm_op( 1, 6, ua_next_long(  ), dt_dword );
	      break;
	    case 11:		// [regl+mem+off],regh
	      reg = ua_next_byte( );
	      regh = ua_next_byte( );
	      disp_op( 0, chunk, ua_next_long(	), reg, 
		  ua_next_long(  ) );
	      reg_op( 1, regh );
	      break;
	    case 12:		// [regl+mem+off],imm
	      reg = ua_next_word( ) & 0xff;
	      disp_op( 0, chunk, ua_next_long(	), reg, 
		  ua_next_long(  ) );
	      if ( chunk <= 1 )
		imm_op( 1, 12, ua_next_word(  ), dt_word );
	      else
		imm_op( 1, 12, ua_next_long(  ), dt_dword );
	      break;
	    case 13:		// [regl+regh],reg2
	      phrase_op( 0, dt_byte, ua_next_byte( ), ua_next_byte(  ) );
	      reg_op( 1, ua_next_word(	) & 0xff );
	      break;
	    case 14:		// [regl+regh],imm
	      phrase_op( 0, dt_byte, ua_next_byte( ), ua_next_byte(  ) );
	      if ( chunk <= 1 )
		imm_op( 1, 4, ua_next_word(  ), dt_word );
	      else
		imm_op( 1, 4, ua_next_long(  ), dt_dword );
	      break;
	  }
	if ( amode == 0xf )
	  switch ( chunk )
	  {
	    case 0:  cmd.itype = KINT_LEA; break;
	    case 1:  cmd.itype = KINT_PEA; break;
	    default: cmd.itype = KINT_UNK; break;
	  }
	else
	  switch ( inst )
	  {
	    case 0: cmd.itype = KINT_LDM; break;
	    case 1: cmd.itype = KINT_ADD; break;
	    case 2: cmd.itype = KINT_SUB; break;
	    case 3: cmd.itype = KINT_CMP; break;
	    case 4: cmd.itype = KINT_TST; break;
	    case 5: cmd.itype = KINT_AND; break;
	    case 6: cmd.itype = KINT_OR;  break;
	    case 7: cmd.itype = KINT_XOR; break;
	  }
      }
      break;
  }
  return cmd.size;
}

//----------------------------------------------------------------------
// analyze an instruction
int	      idaapi
ana( void )
{
  // analyze it!
  return ana_basic(  );
}
