/* Disassembler for KINT virtual processor                              */

#include "kint.hpp"
#include <frame.hpp>
#include <struct.hpp>

static bool   flow;		// does the current instruction pass
				// execution to the next instruction?

extern char  *altCcNames[];

//----------------------------------------------------------------------
// Handle an operand. What this function usually does:
//      - creates cross-references from the operand
//        (the kernel deletes all xrefs before calling emu())
//      - creates permanent comments
//      - if possible, specifies the operand type (for example, it may
//        create stack variables)
//      - anything else you might need to emulate or trace
static void
handle_operand( op_t & x, int loading )
{
  switch ( x.type )
  {
    case o_reg:
    case o_reg_bit:
    case o_phrase:
      break;

    case o_imm:
      // this can't happen!
      if ( !loading )
	goto BAD_LOGIC;

      switch ( cmd.itype )
      {
	case KINT_LDS:
	case KINT_STS:
	  // create stack variables if required
	  if ( may_create_stkvars(  ) && !isDefArg( uFlag, x.n ) )
	  {
	    func_t       *pfn = get_func( cmd.ea );
	    if ( pfn != NULL && pfn->flags & FUNC_FRAME )
	    {
	      struc_t  *s = get_frame( pfn );
	      sval_t        delta;
	      member_t     *mptr = get_stkvar( x, x.value, &delta );
              char 	b[20];
	      if ( mptr == NULL )
	      {
		// set the size according to the op1 register size
		if( cmd.Op1.reg < 0x80 ) 
		  x.dtyp = dt_dword;
		else if( cmd.Op1.reg < 0xe0 )
		  x.dtyp = dt_word;
		else
		  x.dtyp = dt_byte;
		if ( !ua_stkvar2( x, x.value, STKVAR_VALID_SIZE ) )
		  return;
		mptr = get_stkvar( x, x.value, &delta );
		if ( mptr == NULL )
		  return;	// should not happen but better check
	        delta -= pfn->frsize;
	      }
              qsnprintf(b, sizeof b, "var_%X", x.value );
	      set_member_name( s, mptr->soff, b );
	      op_stkvar( cmd.ea, x.n );
	    }
	  }
	  break;
	default:
	  // set immediate flag
	  doImmd( cmd.ea );
	  break;
      }
      // fall through 
    case o_displ:
      // create of delete a data xref if the value was converted to an offset
      {
	xrefblk_t     xb;
	ea_t          pdata = ( x.type == o_displ ) ? x.addr : x.value;
	bool          ok = xb.first_from( cmd.ea, XREF_DATA );

	// msg( "Loc %x xref %d flag %d\n", cmd.ea, ok, isOff(uFlag, x.n) );
	if ( isOff( uFlag, x.n ) && !ok )
	  add_dref( cmd.ea, pdata, dr_O );
	else if ( !isOff( uFlag, x.n ) && ok )
	  del_dref( cmd.ea, pdata );
      }

      break;

    case o_mem:
      ua_dodata2( x.offb, x.addr, x.dtyp );
      doVar( x.addr );
      ua_add_dref( x.offb, x.addr, loading ? dr_R : dr_W );
      switch ( cmd.itype )
      {
	case KINT_JSI:		// add a code reference for indirect calls
	case KINT_JSIC:
	  ua_add_cref( x.offb, get_long( x.addr ), fl_CN );
	  break;
      }

      break;

    case o_near:
      {
	// work out if it is a CALL, and add in a code xref
	bool          iscall = InstrIsSet( cmd.itype, CF_CALL );
	ua_add_cref( x.offb, x.addr, iscall ? fl_CN : fl_JN );

	// if dest is a non-returning function, don't flow onto next op
	if ( flow && iscall )
	{
	  if ( !func_does_return( x.addr ) )
	    flow = false;
	}
	break;
      }

    default:
    BAD_LOGIC:
      warning( "%a (%s): bad optype", cmd.ea, cmd.get_canon_mnem(  ) );
      break;
  }
}

//----------------------------------------------------------------------
// Emulate an instruction
// This function should:
//      - create all xrefs from the instruction
//      - perform any additional analysis of the instruction/program
//        and convert the instruction operands, create comments, etc.
//      - create stack variables
//      - analyze the delayed branches and similar constructs
// The kernel calls ana() before calling emu(), so you may be sure that
// the 'cmd' structure contains a valid and up-to-date information.
// You are not allowed to modify the 'cmd' structure.
// Upon entering this function, the 'uFlag' variable contains the flags of
// cmd.ea. If you change the characteristics of the current instruction, you
// are required to refresh 'uFlag'.
// Usually the kernel calls emu() with consecutive addresses in cmd.ea but
// you can't rely on this - for example, if the user asks to analyze an
// instruction at arbirary address, his request will be handled immediately,
// thus breaking the normal sequence of emulation.
// If you need to analyze the surroundings of the current instruction, you
// are allowed to save the contents of the 'cmd' structure and call ana().
// For example, this is a very common pattern:
//  {
//    insn_t saved = cmd;
//    if ( decode_prev_insn(cmd.ea) != BADADDR )
//    {
//      ....
//    }
//    cmd = saved;
//  }
//
// This sample emu() function is a very simple emulation engine.

int           idaapi
emu( void )
{
  // setup
  ulong         Feature = cmd.get_canon_feature(  );
  flow = true;
  int		offdiff;
  uint          jstr, i;
  ea_t          eaddr;

  // disable flow if CF_STOP set 
  if ( Feature & CF_STOP ) 
  {
    // unless next instruction is a PUSHC
    if( get_word( cmd.ea + cmd.size ) != 0xfe63 ) 
      flow = false;
  }
  // deal with operands
  if ( Feature & CF_USE1 ) handle_operand( cmd.Op1, 1 );
  if ( Feature & CF_USE2 ) handle_operand( cmd.Op2, 1 );
  if ( Feature & CF_JUMP ) QueueMark( Q_jumps, cmd.ea );
  if ( Feature & CF_CHG1 ) handle_operand( cmd.Op1, 0 );
  if ( Feature & CF_CHG2 ) handle_operand( cmd.Op2, 0 );

  // if the execution flow is not stopped here, then create
  // a xref to the next instruction.
  // Thus we plan to analyze the next instruction.
  if ( flow )
    ua_add_cref( 0, cmd.ea + cmd.size, fl_F );

  switch ( cmd.itype )
  {
    case KINT_JMP:
      if( cmd.Op1.type != o_displ )
        break;
      if( ( jstr = cmd.Op1.addr ) == 0 )
	break;
      offdiff = cmd.ea + cmd.size - jstr;
      if( offdiff > 40 * 4 || offdiff < -2 || offdiff & 1 )
	break;
      eaddr = cmd.ea + cmd.size;
      if( (offdiff & 3) == 2 )
      {
	doWord( eaddr, 2 );
	eaddr += 2;
      }
      {
	xrefblk_t     xb;
	bool          ref = xb.first_from( cmd.ea, XREF_DATA );
        if( !ref) 
	{
	  add_dref( cmd.ea, eaddr, dr_O );
	  if( offdiff <= 0 )
            set_offset( cmd.ea, 0, 0 );
	}
        set_offset( eaddr, 0, 0 );    // at leas one entry
        jstr = get_long( eaddr );
        auto_make_code( jstr );
	eaddr += 4;
        for( i = 0; i < 40 - 1; i++ ) // maximum 40 jump entries
        {
	  ref = xb.first_from( eaddr, XREF_ALL );
	  if( ref ) 
	    break;
          jstr = get_long( eaddr );
          auto_make_code( jstr );
	  eaddr += 4;
        }
      }
      break;
    case KINT_IJMP:
      jstr = cmd.Op2.addr;
      doDwrd( jstr, 4 );
      for ( i = 0; i < get_byte( jstr ); i++ )
      {
	eaddr = jstr + 4 * ( i + 1 );
	set_offset( eaddr, 0, 0 );
	eaddr = get_long( eaddr );
	auto_make_code( eaddr );
      }
      break;
    case KINT_KJMP:
      jstr = cmd.Op2.addr;
      doDwrd( jstr, 4 );
      for ( i = 0; i < get_byte( jstr ); i++ )
      {
	eaddr = jstr + 4 + 8 * i;
	doDwrd( eaddr, 4 );
	set_offset( eaddr + 4, 0, 0 );
	eaddr = get_long( eaddr + 4 );
	auto_make_code( eaddr );
      }
      break;
  }

  // OK (actual code unimportant)
  return 1;
}

//----------------------------------------------------------------------
bool
create_func_frame( func_t * pfn )
{
  if ( pfn != NULL )
  {
    if ( pfn->frame == BADNODE )
    {
      ea_t          ea = pfn->startEA;
      if ( ea + 4 < pfn->endEA )	// minimum 4 bytes needed
      {
	decode_insn( ea );
	if ( cmd.itype == KINT_AFR )
	{
	  size_t        localsize = ( size_t ) cmd.Op1.value;
	  decode_insn( ea + 4 );
	  setflag( ( uint32 & ) pfn->flags, FUNC_FRAME, 1 );
	  return add_frame( pfn, localsize, 0, 0 );
	}
      }
    }
  }
  return 0;
}

