/* Disassembler for KINT virtual processor				*/

#include "kint.hpp"

// CF_STOP 0x00001  // Instruction doesn't pass execution to the
//		    // next instruction
// CF_CALL 0x00002  // CALL instruction (should make a procedure here)
// CF_CHG1 0x00004  // The instruction modifies the first operand
// CF_CHG2 0x00008  // The instruction modifies the second operand
// CF_USE1 0x00100  // The instruction uses value of the first operand
// CF_USE2 0x00200  // The instruction uses value of the second operand
// CF_JUMP 0x04000  // The instruction passes execution using indirect
//		    // jump or call (thus needs additional analysis)
// CF_SHFT 0x08000  // Bit-shift instruction (shl,shr...)
// CF_HLL  0x10000  // Instruction may be present in a high level
//		    // language function.

instruc_t Instructions[] = {

{ "",		0		},	// Unknown Operation

{ "abit",	0		},
{ "add",	CF_USE1|CF_USE2	},
{ "adf",	0		},
{ "afr", 	0		},
{ "and",	0		},
{ "b2i",	0		},
{ "brk",	0		},
{ "cbd",	0		},
{ "cbw",	0		},
{ "clr",	0		},
{ "cmf",	0		},
{ "cmp",	CF_USE1|CF_USE2	},
{ "cwd",	0		},
{ "dec",	0		},
{ "dfr",	0		},
{ "dif",	0		},
{ "div",	0		},
{ "f2i",	CF_USE1		},
{ "fadd",	0		},
{ "fcmp",	0		},
{ "fdiv",	0		},
{ "ff2i",	CF_USE1		},
{ "fi2f",	CF_USE1		},
{ "fmul",	0		},
{ "fneg",	0		},
{ "fsub",	0		},
{ "HALT",	CF_STOP		},
{ "i2f",	CF_USE1		},
{ "ijmp",	CF_JUMP|CF_USE2 },
{ "inc",	0		},
{ "je",		CF_USE1		},
{ "jev",	CF_USE1		},
{ "jg",		CF_USE1		},
{ "jge",	CF_USE1		},
{ "jl",		CF_USE1		},
{ "jle",	CF_USE1		},
{ "jmp",	CF_USE1|CF_STOP	}, // short jump
{ "jmp",	CF_USE1|CF_STOP	}, // long jump
{ "jne",	CF_USE1		},
{ "jod",	0		},
{ "jsi",	CF_USE1|CF_CALL	},
{ "jsic",	CF_USE1|CF_CALL	},
{ "jsr",	CF_USE1|CF_CALL	},
{ "jsrc",	CF_USE1|CF_CALL	},
{ "kjmp",	CF_JUMP|CF_USE2 },
{ "ld0",	CF_CHG1|CF_USE2	},
{ "ld1",	CF_CHG1|CF_USE2	},
{ "ld2",	CF_CHG1|CF_USE2	},
{ "ldc",	0		},
{ "ldcm",	0		},
{ "ldcp",	0		},
{ "lde",	0		},
{ "ldi",	CF_CHG1|CF_USE2	},
{ "ldm",	CF_USE1|CF_USE2	},
{ "ldr",	0		},
{ "lds",	CF_USE1|CF_USE2	},
{ "lea",	CF_USE1|CF_USE2	},
{ "lje",	CF_USE1		},
{ "ljev",	CF_USE1		},
{ "ljg",	CF_USE1		},
{ "ljge",	CF_USE1		},
{ "ljl",	CF_USE1		},
{ "ljle",	CF_USE1		},
{ "ljne",	CF_USE1		},
{ "ljod",	CF_USE1		},
{ "lsa",	0		},
{ "mod",	0		},
{ "muf",	0		},
{ "mul",	0		},
{ "nef",	0		},
{ "neg",	0		},
{ "nop",	0		},
{ "not",	0		},
{ "or",		0		},
{ "pea",	CF_USE1		},
{ "po0",	0		},
{ "pof",	0		},
{ "pop",	CF_USE1		},
{ "popc",	CF_USE1		},
{ "por",	0		},
{ "pu0",	0		},
{ "pub",	CF_USE1		},
{ "pud",	CF_USE1		},
{ "puf",	0		},
{ "pur",	0		},
{ "pusc",	CF_USE1		},
{ "push",	CF_USE1		},
{ "puw",	CF_USE1		},
{ "ref",	0		},
{ "ret",	CF_STOP		},
{ "retc",	CF_STOP		},
{ "rol",	CF_SHFT		},
{ "ror",	CF_SHFT		},
{ "saf",	0		},
{ "shl",	CF_SHFT		},
{ "shr",	CF_SHFT		},
{ "sm",		0		},
{ "st0",	CF_USE1|CF_USE2	},
{ "st1",	CF_USE1|CF_USE2	},
{ "st2",	CF_USE1|CF_USE2	},
{ "stc",	0		},
{ "stcm",	0		},
{ "stcp",	0		},
{ "sts",	CF_USE1|CF_USE2	},
{ "sub",	CF_USE1|CF_USE2	},
{ "suf",	0		},
{ "TERM",	CF_STOP		},
{ "tst",	0		},
{ "usm",	0		},
{ "xch",	0		},
{ "xor",	0		},

{ "SysFileSize",	CF_HLL		},
{ "SysOpen",		CF_HLL		},
{ "SysEof",		CF_HLL		},
{ "SysRead",		CF_HLL		},
{ "SysWrite",		CF_HLL		},
{ "SysSeek",		CF_HLL		},
{ "SysClose",		CF_HLL		},
{ "SysFindFirst",	CF_HLL		},
{ "SysFindNext",	CF_HLL		},
{ "SysFindClose",	CF_HLL		},
{ "SysDelete",		CF_HLL		},
{ "SysMakeShareCopy",	CF_HLL		},
{ "SysPrintStart",	CF_HLL		},
{ "SysPrintBegPage",	CF_HLL		},
{ "SysPrintEndPage",	CF_HLL		},
{ "SysPrintStop",	CF_HLL		},
{ "SysAllocMem",	CF_HLL		},
{ "SysFreeMem",		CF_HLL		},
{ "SysTimeReg",		CF_HLL		},
{ "SysTimeStr",		CF_HLL		},
{ "SysTimeMilli",	CF_HLL		},
{ "SysStrCat",		CF_HLL		},
{ "SysMemCpy",		CF_HLL		},
{ "SysMemMove",		CF_HLL		},
{ "SysStrCpy",		CF_HLL		},
{ "SysStrCmp",		CF_HLL		},
{ "SysStrICmp",		CF_HLL		},
{ "SysStrLen",		CF_HLL		},
{ "SysMemSet",		CF_HLL		},
{ "SysMemCmp",		CF_HLL		},
{ "SysSetCursor",	CF_HLL		},
{ "SysSetPGMTitle",	CF_HLL		},
{ "SysCreBM",		CF_HLL		},
{ "SysDelBM",		CF_HLL		},
{ "SysAlphaBlend",	CF_HLL		},
{ "SysSizeBM",		CF_HLL		},
{ "SysDrawBM",		CF_HLL		},
{ "SysStretchBM",	CF_HLL		},
{ "SysDrawMBM",		CF_HLL		},
{ "SysLoadBM",		CF_HLL		},
{ "SysLoadMBM",		CF_HLL		},
{ "SysStretchMBM",	CF_HLL		},
{ "SysGetWindowPos",	CF_HLL		},
{ "SysSetWindowPos",	CF_HLL		},
{ "SysGetViewSize",	CF_HLL		},
{ "SysOrientBM",	CF_HLL		},
{ "SysGetScreenSize",	CF_HLL		},
{ "SysFlushScreen",	CF_HLL		},
{ "SysRectEdged",	CF_HLL		},
{ "SysRectSolid",	CF_HLL		},
{ "SysRectOutline",	CF_HLL		},
{ "SysLines",		CF_HLL		},
{ "SysSetPixel",	CF_HLL		},
{ "SysGetPixel",	CF_HLL		},
{ "SysColorize",	CF_HLL		},
{ "SysRect",		CF_HLL		},
{ "SysOval",		CF_HLL		},
{ "SysGetTextSize",	CF_HLL		},
{ "SysGetFont",		CF_HLL		},
{ "SysRelFont",		CF_HLL		},
{ "SysBegText",		CF_HLL		},
{ "SysTextStyle",	CF_HLL		},
{ "SysText",		CF_HLL		},
{ "SysEndText",		CF_HLL		},
{ "SysPlaySound",	CF_HLL		},
{ "SysPlayMusic",	CF_HLL		},
{ "SysStopMusic",	CF_HLL		},
{ "SysIsMusic",		CF_HLL		},
{ "SysCaptureMouse",	CF_HLL		},
{ "SysMemSET2",		CF_HLL		},
{ "SysMemSET4",		CF_HLL		},
{ "SysPrintFONT",	CF_HLL		},
{ "SysStartTimer",	CF_HLL		},
{ "SysStopTimer",	CF_HLL		},
{ "SysGetEOL",		CF_HLL		},
{ "SysPrintGetTextSize", CF_HLL		},
{ "SysPrintText",	CF_HLL		},
{ "SysPrintBM",		CF_HLL		},
{ "SysPrintLines",	CF_HLL		},
{ "SysSystemMSG",	CF_HLL		},
{ "SysLoadKE",		CF_HLL		},
{ "SysUnloadKE",	CF_HLL		},
{ "SysMsgKE",		CF_HLL		},
{ "Sys0x53",		CF_HLL		},
{ "SysLoadLossyBM",	CF_HLL		},
{ "SysInvDrawBM",	CF_HLL		},
{ "SysInvDrawMBM",	CF_HLL		},
{ "SysLineStyle",	CF_HLL		},
{ "SysMakeFolder",	CF_HLL		},
{ "SysDelFolder",	CF_HLL		},
{ "SysClip",		CF_HLL		},
{ "SysGetClip",		CF_HLL		},
{ "SysMaskBM",		CF_HLL		},
{ "SysRGB2HSL",		CF_HLL		},
{ "SysHSL2RGB",		CF_HLL		},
{ "SysSaveBM",		CF_HLL		},
{ "SysPrintRect",	CF_HLL		},
{ "SysSleep",		CF_HLL		},
{ "SysGetSystemData",	CF_HLL		},
{ "SysMemSet200",	CF_HLL		},
{ "SysMemSet400",	CF_HLL		},
{ "SysStrCat00",	CF_HLL		},
{ "SysMemCpy00",	CF_HLL		},
{ "SysMemMove00",	CF_HLL		},
{ "SysStrCpy00",	CF_HLL		},
{ "SysStrCmp00",	CF_HLL		},
{ "SysStrICmp00",	CF_HLL		},
{ "SysStrLen00",	CF_HLL		},
{ "SysMemSet00",	CF_HLL		},
{ "SysMemCmp00",	CF_HLL		},
{ "UnknownInst",	0		},
  };

#ifdef __BORLANDC__
#if sizeof(Instructions)/sizeof(Instructions[0]) != KINT_last
#error		No match:  sizeof(InstrNames) !!!
#endif
#endif
