#include "kint.hpp"
#include <fpro.h>
#include <diskio.hpp>
#include <struct.hpp>
#pragma warning(disable: 4800)


//----------------------------------------------------------------------
void OutRegString( int regNum, int regBit = -1) {
  char buf[256];

  if( regNum < 0x80 )
    regNum /= 4;
  else if( regNum < 0xe0 )
    regNum = 32 + (regNum - 0x80) / 2;
  else
    regNum = 80 + regNum - 0xe0;
  qsnprintf(buf, sizeof(buf), "r%u", regNum );
  out_register(buf);

  // output regBit if requested
  if (regBit != -1) {
    qsnprintf(buf, sizeof(buf), ".%i", regBit);
    out_line(buf, COLOR_DEFAULT);
  }
}

//----------------------------------------------------------------------
void OutAddr(op_t& x, ea_t ea, ea_t off, bool isSigned = false) {
  // try and find the real name expression
  if ( !out_name_expr(x, ea, off) )
  {
    // work out flags correctly
    ulong flags = OOF_ADDR | OOFW_16;
    if (isSigned) flags |= OOF_SIGNED;
    else flags |= OOFS_NOSIGN;

    // if name wasn't found, just output the value & add to noname queue
    OutValue(x, flags);
    QueueMark(Q_noName, cmd.ea);
  }
}

void idaapi out_itype( op_t &x) {
  switch( x.dtyp )
  {
    case dt_byte:
      out_symbol('B');
      break;
    case dt_word:
      out_symbol('W');
      break;
    case dt_dword:
      out_symbol('D');
      break;
  }
}


//----------------------------------------------------------------------
// generate the text representation of an operand

bool idaapi outop(op_t &x) {
  // output operands
  switch(x.type) {
  case o_reg:
    OutRegString( x.reg );
    break;

  case o_reg_bit:
    OutRegString( x.reg, x.v_bit);
    break;

  case o_phrase:
    out_symbol('[');
    OutRegString( x.phrase );
    if( x.specval < 0xffff )
    {
      out_symbol('+');
      OutRegString( x.specval );
    }
    out_symbol(']');
    out_itype(x);
    break;

  case o_displ:
    out_symbol('[');
    if( x.phrase < 0xffff )
    {
      OutRegString( x.phrase );
      out_symbol('+');
    }
    if( x.value )
    {
      OutValue(x, OOFS_IFSIGN | OOF_SIGNED | OOFW_IMM);
      out_symbol('+');
    }
    OutValue(x, OOF_ADDR | OOFS_IFSIGN | OOF_SIGNED | OOFW_IMM);
    out_symbol(']');
    out_itype(x);
    break;

  case o_imm:
    if( cmd.itype != KINT_POF )
      out_symbol('=');
    OutValue(x, OOFS_IFSIGN | OOF_SIGNED | OOFW_IMM);
    break;

  case o_near:
  case o_mem:
    switch( cmd.itype ) // add a '=' for calls, jumps and case jumps
    {
      case KINT_JSR:
      case KINT_JSRC:
      case KINT_LJE:
      case KINT_LJNE:
      case KINT_LJL:
      case KINT_LJLE:
      case KINT_LJG:
      case KINT_LJGE:
      case KINT_LJEV:
      case KINT_LJOD:
      case KINT_LJMP:
      case KINT_IJMP:
      case KINT_KJMP:
        out_symbol('=');
        break;
    }
    OutAddr(x, x.addr, x.addr);
    break;

  }

  // OK
  return 1;
}

//----------------------------------------------------------------------
// generate a text representation of an instruction
// the information about the instruction is in the 'cmd' structure
void idaapi out(void) {
  char buf[MAXSTR];

  // setup the output pointer
  init_output_buffer(buf, sizeof(buf));

  switch( cmd.itype )
  {
    case KINT_AFR:
      out_insert( get_output_ptr( ), COLSTR("@frame_end   ; ",SCOLOR_LOCNAME) );
      break;
    case KINT_DFR:
      out_insert( get_output_ptr( ), COLSTR("@frame_del",SCOLOR_LOCNAME) );
      break;
    default:
      // output instruction mnemonics
      OutMnem();
      break;
  }

  // output the first operand
  if (cmd.Op1.type != o_void) {
    out_one_operand(0);
  }

  // output the second operand
  if (cmd.Op2.type != o_void) {
    out_symbol(',');
    OutChar(' ');
    out_one_operand(1);
  }

  // output the third operand
  if (cmd.Op3.type != o_void) {
    out_symbol(',');
    OutChar(' ');
    out_one_operand(2);
  }

  // terminate the output string
  term_output_buffer();

  // ask to attach a possible user-defined comment to it
  gl_comm = 1;

  // pass the generated line to the kernel
  MakeLine(buf);
}


//--------------------------------------------------------------------------
// generate start of the disassembly
void idaapi header(void) {
  // generate standard header
  gen_cmt_line("Processor:	  %s", inf.procName);
  gen_cmt_line("Target assembler: %s", ash.name);

  // output assembler-specific header
  if (ash.header != NULL) {
    for (const char **ptr=ash.header; *ptr != NULL; ptr++) {
      MakeLine(*ptr, 0);
    }
  }
}


// --------------------------------------------------------------------------
// generate start of segment
void idaapi segstart(ea_t ea) {
  char sgmt_name[MAXSTR];
  // get segment data
  segment_t *Sarea = getseg(ea);

  // generate ORG directive if necessary
  if (inf.s_org)
  {
    ulong org = ea - get_segm_base(Sarea);
    // generate line
    if ( org != 0 )
    {
      char buf[MAX_NUMBUF];
      btoa(buf, sizeof(buf), org);
      printf_line(inf.indent, COLSTR("%s %s", SCOLOR_ASMDIR), ash.origin, buf);
    }
  }
  sgmt_name[0] = '@';
  get_segm_name(Sarea, sgmt_name+1, MAXSTR - 2);
  printf_line(0, COLSTR("%s", SCOLOR_ASMDIR), sgmt_name );
  if( ea == 0 )
    printf_line(0, COLSTR("%s", SCOLOR_ASMDIR), "@align off" );
}


// --------------------------------------------------------------------------
// generate end of segment
void idaapi segfooter(ea_t ea) {
  char sgmt_name[MAXSTR];
  // get segment data
  segment_t *Sarea = getseg(ea-1);

  sgmt_name[0] = '@';
  get_segm_name(Sarea, sgmt_name+1, MAXSTR - 2);
  qstrncat( sgmt_name, "_end", MAXSTR - 2 );
  printf_line(0, COLSTR("%s", SCOLOR_ASMDIR), sgmt_name );
}

// --------------------------------------------------------------------------
// generate end of the disassembly
void idaapi footer(void) {
  char buf[MAXSTR];

  // if assembler supplies end statement, output it
  if (ash.end != NULL) {
    MakeNull();
    tag_addstr(buf, buf+sizeof(buf), COLOR_ASMDIR, ash.end);
    MakeLine(buf, inf.indent);
  }
}

// --------------------------------------------------------------------------
// customised address output
void idaapi out_data(ea_t ea) {
  // if addres is valid, use normal output function
  if (1 || isLoaded(ea)) {
    intel_data(ea);
  } else {
    gl_name = 1;
    MakeLine(COLSTR("BYTE ?", SCOLOR_SYMBOL));
  }
}

//--------------------------------------------------------------------------
// prints function header
void kint_func_header(func_t *pfn)
{
  char buf[MAXSTR];

  std_gen_func_header(pfn);
  get_func_name( pfn->startEA, buf, MAXSTR );
  printf_line(0, COLSTR( "\3%s:", SCOLOR_CNAME ), buf );
  if( pfn->flags & FUNC_FRAME )
  {
    printf_line(inf.indent, COLSTR("@frame_args  ; Start Frame" , SCOLOR_LOCNAME ) );
    printf_line(inf.indent, COLSTR("@frame_vars", SCOLOR_LOCNAME ) );
  }
}


//--------------------------------------------------------------------------
void gen_stkvar_def(char *buf, size_t bufsize, const member_t *mptr, sval_t v)
{
  char name[MAXNAMELEN];
  const char *tp;
  int size = mptr->eoff - mptr->soff;

  name[0] = '\0';
  switch( size )
  {
    case 1: tp = "BYTE";  break;
    case 2: tp = "WORD";  break;
    case 4: tp = "DWORD"; break;
  }
  get_member_name(mptr->id, name, sizeof(name));
  qsnprintf(buf, bufsize, COLSTR("%*c%s:%*s",SCOLOR_LOCNAME),
               inf.indent, ' ', name, 16 - strlen( name),  tp );
}
